# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015, 2016 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Multimedia Image API Tests."""

from io import BytesIO

import pytest
from PIL import Image

from flask_iiif.api import MultimediaImage
from flask_iiif.utils import create_gif_from_frames

from .helpers import IIIFTestCase


class TestMultimediaAPI(IIIFTestCase):
    """Multimedia Image API test case."""

    @pytest.mark.parametrize("width, height",
                             [(1280, 1024),
                              (100, 100),
                              (1024, 1280),  # portrait
                              (200, 100),
                              (1280, 720),
                              ])
    def setUp(self, width=1280, height=1024):
        """Run before the test."""
        # Create an image in memory

        tmp_file = BytesIO()
        # create a new image
        image = Image.new("RGBA", (width, height), (255, 0, 0, 0))
        image.save(tmp_file, 'png')

        # create a new gif image
        self.image_gif = MultimediaImage(create_gif_from_frames([
            Image.new("RGB", (width, height), color)
            for color in ['blue', 'yellow', 'red', 'black', 'white']
        ]))
        self.width, self.height = width, height
        # Initialize it for our object and create and instance for
        # each test
        tmp_file.seek(0)
        self.image_resize = MultimediaImage.from_string(tmp_file)
        tmp_file.seek(0)
        self.image_crop = MultimediaImage.from_string(tmp_file)
        tmp_file.seek(0)
        self.image_rotate = MultimediaImage.from_string(tmp_file)
        tmp_file.seek(0)
        self.image_errors = MultimediaImage.from_string(tmp_file)
        tmp_file.seek(0)
        self.image_convert = MultimediaImage.from_string(tmp_file)
        tmp_file.seek(0)
        self.image_save = MultimediaImage.from_string(tmp_file)

        # NOT RGBA
        tmp_file = BytesIO()
        # create a new image
        image = Image.new("RGB", (width, height), (255, 0, 0, 0))
        image.save(tmp_file, 'jpeg')
        tmp_file.seek(0)
        self.image_not_rgba = MultimediaImage.from_string(tmp_file)

        # Image in P Mode
        tmp_file = BytesIO()
        image = Image.new("P", (width, height))
        image.save(tmp_file, 'gif')
        tmp_file.seek(0)
        self.image_p_mode = MultimediaImage.from_string(tmp_file)

        # TIFF Image
        tmp_file = BytesIO()
        image = Image.new("RGBA", (width, height), (255, 0, 0, 0))
        image.save(tmp_file, 'tiff')
        tmp_file.seek(0)
        self.image_tiff = MultimediaImage.from_string(tmp_file)

    def test_gif_resize(self):
        """Test image resize function on GIF images."""
        # Check original size and GIF properties
        self.assertEqual(self.image_gif.image.is_animated, True)
        self.assertEqual(self.image_gif.image.n_frames, 5)
        self.assertEqual(str(self.image_gif.size()),
                         str((self.width, self.height)))

        # Assert proper resize and preservation of GIF properties
        self.image_gif.resize('720,680')
        self.assertEqual(self.image_gif.image.is_animated, True)
        self.assertEqual(self.image_gif.image.n_frames, 5)
        self.assertEqual(str(self.image_gif.size()), str((720, 680)))

        # test gif resize best fit
        self.image_gif.resize('!400,220')
        self.assertEqual(self.image_gif.image.is_animated, True)
        self.assertEqual(self.image_gif.image.n_frames, 5)
        self.assertEqual(str(self.image_gif.size()), str((400, 220)))

    def test_image_resize(self):
        """Test image resize function."""
        # Test image size before
        self.assertEqual(str(self.image_resize.size()),
                         str((self.width, self.height)))

        # Resize.image_resize to 720,680
        self.image_resize.resize('720,680')
        self.assertEqual(str(self.image_resize.size()), str((720, 680)))

        # Resize.image_resize to 300,
        self.image_resize.resize('300,')
        self.assertEqual(str(self.image_resize.size()), str((300, 283)))

        # Resize.image_resize to ,300
        self.image_resize.resize(',300')
        self.assertEqual(str(self.image_resize.size()), str((318, 300)))

        # Resize.image_resize to pct:90
        self.image_resize.resize('pct:90')
        self.assertEqual(str(self.image_resize.size()), str((286, 270)))

        # Resize.image_resize to !100,100
        self.image_resize.resize('!100,100')
        self.assertEqual(str(self.image_resize.size()), str((100, 100)))

    def test_errors(self):
        """Test errors."""
        from flask_iiif.errors import (
            MultimediaImageResizeError, MultimediaImageCropError,
            MultimediaImageNotFound, MultimediaImageQualityError,
            MultimediaImageFormatError
        )
        # Test resize errors
        self.assertRaises(
            MultimediaImageResizeError,
            self.image_errors.resize,
            'pct:-12222'
        )
        self.assertRaises(
            MultimediaImageResizeError,
            self.image_errors.resize,
            '2'
        )
        self.assertRaises(
            MultimediaImageResizeError,
            self.image_errors.resize,
            '-22,100'
        )
        # Test crop errors
        self.assertRaises(
            MultimediaImageCropError,
            self.image_errors.crop,
            '22,100,222'
        )
        self.assertRaises(
            MultimediaImageCropError,
            self.image_errors.crop,
            '-22,100,222,323'
        )
        self.assertRaises(
            MultimediaImageCropError,
            self.image_errors.crop,
            'pct:222,100,222,323'
        )
        self.assertRaises(
            MultimediaImageCropError,
            self.image_errors.crop,
            '2000,2000,2000,2000'
        )
        self.assertRaises(
            MultimediaImageNotFound,
            self.image_errors.from_file,
            "unicorn"
        )
        with self.app.app_context():
            self.assertRaises(
                MultimediaImageQualityError,
                self.image_errors.quality,
                'pct:222,100,222,323'
            )
            self.assertRaises(
                MultimediaImageFormatError,
                self.image_errors._prepare_for_output,
                'pct:222,100,222,323'
            )

    def test_image_crop(self):
        """Test the crop function."""
        # Crop image x,y,w,h
        self.image_crop.crop('20,20,400,300')
        self.assertEqual(str(self.image_crop.size()), str((400, 300)))

        # Crop image pct:x,y,w,h
        self.image_crop.crop('pct:20,20,40,30')
        self.assertEqual(str(self.image_crop.size()), str((160, 90)))

        # Check if exeeds image borders
        self.image_crop.crop('10,10,160,90')
        self.assertEqual(str(self.image_crop.size()), str((150, 80)))

    def test_image_rotate(self):
        """Test image rotate function."""
        self.image_rotate.rotate(90)
        self.assertEqual(str(self.image_rotate.size()), str((1024, 1280)))

        self.image_rotate.rotate(120)
        self.assertEqual(str(self.image_rotate.size()), str((1024, 1280)))

    def test_image_mode(self):
        """Test image mode."""
        self.image_not_rgba.quality('grey')
        self.assertEqual(self.image_not_rgba.image.mode, "L")

    def test_image_incompatible_modes(self):
        """Test P-incompatible image to RGB auto-convert."""
        tmp_file = BytesIO()
        self.image_p_mode.save(tmp_file)
        self.assertEqual(self.image_p_mode.image.mode, "RGB")

    def test_image_saving(self):
        """Test image saving."""
        tmp_file = BytesIO()
        compare_file = BytesIO()
        self.assertEqual(tmp_file.getvalue(), compare_file.getvalue())
        self.image_save.save(tmp_file)
        self.assertNotEqual(str(tmp_file.getvalue()), compare_file.getvalue())

    def test_image_tiff_support(self):
        """Test TIFF image support."""
        self.assertEqual(self.image_tiff.image.format, 'TIFF')



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015, 2016, 2017 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Multimedia Image API."""

import itertools
import math
import os
import re

from flask import current_app
from PIL import Image
from six import BytesIO, string_types
from werkzeug import import_string

from .errors import IIIFValidatorError, MultimediaImageCropError, \
    MultimediaImageFormatError, MultimediaImageNotFound, \
    MultimediaImageQualityError, MultimediaImageResizeError, \
    MultimediaImageRotateError
from .utils import fill_background, resize_gif, resize_with_background_gif


class MultimediaObject(object):
    """The Multimedia Object."""


class MultimediaImage(MultimediaObject):
    r"""Multimedia Image API.

    Initializes an image api with IIIF standards. You can:

    * Resize :func:`resize`.
    * Crop :func:`crop`.
    * Rotate :func:`rotate`.
    * Change image quality :func:`quality`.

    Example of editing an image and saving it to disk:

    .. code-block:: python

        from flask_iiif.api import MultimediaImage

        image = IIIFImageAPIWrapper.from_file(path)
        # Rotate the image
        image.rotate(90)
        # Resize the image
        image.resize('300,200')
        # Crop the image
        image.crop('20,20,400,300')
        # Make the image black and white
        image.quality('grey')
        # Finaly save it to /tmp
        image.save('/tmp')

    Example of serving the modified image over http:

    .. code-block:: python

        from flask import current_app, Blueprint
        from flask_iiif.api import MultimediaImage

        @blueprint.route('/serve/<string:uuid>/<string:size>')
        def serve_thumbnail(uuid, size):
            \"\"\"Serve the image thumbnail.

            :param uuid: The document uuid.
            :param size: The desired image size.
            \"\"\"
            # Initialize the image with the uuid
            path = current_app.extensions['iiif'].uuid_to_path(uuid)
            image = IIIFImageAPIWrapper.from_file(path)
            # Resize it
            image.resize(size)
            # Serve it
            return send_file(image.serve(), mimetype='image/jpeg')
    """

    def __init__(self, image):
        """Initialize the image."""
        self.image = image

    @classmethod
    def from_file(cls, path):
        """Return the image object from the given path.

        :param str path: The absolute path of the file
        :returns: a :class:`~flask_iiif.api.MultimediaImage`
                  instance
        """
        if not os.path.exists(path):
            raise MultimediaImageNotFound(
                "The requested image {0} not found".format(path))

        image = Image.open(path)
        return cls(image)

    @classmethod
    def from_string(cls, source):
        """Create an :class:`~flask_iiif.api.MultimediaImage` instance.

        :param str source: the image string
        :type source: `BytesIO` object
        :returns: a :class:`~flask_iiif.api.MultimediaImage`
                  instance
        """
        image = Image.open(source)
        return cls(image)

    def resize(self, dimensions, resample=None):
        """Resize the image.

        :param str dimensions: The dimensions to resize the image
        :param resample: The algorithm to be used
        :type resample: :py:mod:`PIL.Image` algorithm

        .. note::

            * `dimensions` must be one of the following:

                * 'w,': The exact width, height will be calculated.
                * ',h': The exact height, width will be calculated.
                * 'pct:n': Image percentage scale.
                * 'w,h': The exact width and height.
                * '!w,h': Best fit for the given width and height.

        """
        real_width, real_height = self.image.size
        fits, point_x, point_y = True, 0, 0
        if resample is None:
            if isinstance(current_app.config['IIIF_RESIZE_RESAMPLE'],
                          string_types):
                resample = import_string(
                    current_app.config['IIIF_RESIZE_RESAMPLE'])
            else:
                resample = current_app.config['IIIF_RESIZE_RESAMPLE']

        # Check if it is `pct:`
        if dimensions.startswith('pct:'):
            percent = float(dimensions.split(':')[1]) * 0.01
            if percent < 0:
                raise MultimediaImageResizeError(
                    ("Image percentage could not be negative, {0} has been"
                     " given").format(percent)
                )

            width = max(1, int(real_width * percent))
            height = max(1, int(real_height * percent))

        # Check if it is `,h`
        elif dimensions.startswith(','):
            height = int(dimensions[1:])
            # find the ratio
            ratio = self.reduce_by(height, real_height)
            # calculate width (minimum 1)
            width = max(1, int(real_width * ratio))

        # Check if it is `!w,h`
        elif dimensions.startswith('!'):
            point_x, point_y = map(int, dimensions[1:].split(','))
            # find the ratio
            ratio_x = self.reduce_by(point_x, real_width)
            ratio_y = self.reduce_by(point_y, real_height)
            # take the min
            ratio = min(ratio_x, ratio_y)
            # calculate the dimensions
            width = max(1, int(real_width * ratio))
            height = max(1, int(real_height * ratio))
            # check if it's not too small for the requested window
            if not (width == point_x and height == point_y):
                fits = False

        # Check if it is `w,`
        elif dimensions.endswith(','):
            width = int(dimensions[:-1])
            # find the ratio
            ratio = self.reduce_by(width, real_width)
            # calculate the height
            height = max(1, int(real_height * ratio))

        # Normal mode `w,h`
        else:
            try:
                width, height = map(int, dimensions.split(','))
            except ValueError:
                raise MultimediaImageResizeError(
                    "The request must contain width,height sequence"
                )

        # If a dimension is missing throw error
        if any((dimension <= 0 and
                dimension is not None) for dimension in (width, height)):
            raise MultimediaImageResizeError(
                ("Width and height cannot be zero or negative, {0},{1} has"
                 " been given").format(width, height)
            )

        arguments = dict(size=(width, height), resample=resample)
        if self.image.format == 'GIF':
            if not fits:
                self.image = resize_with_background_gif(
                    self.image, demand_w=point_x, demand_h=point_y,
                    **arguments)
            else:
                self.image = resize_gif(self.image, **arguments)
        else:
            if fits:
                self.image = self.image.resize(**arguments)
            else:
                self.image = fill_background(self.image.resize(**arguments),
                                             point_x, point_y)

    def crop(self, coordinates):
        """Crop the image.

        :param str coordinates: The coordinates to crop the image

        .. note::

            * `coordinates` must have the following pattern:

                * 'x,y,w,h': in pixels.
                * 'pct:x,y,w,h': percentage.

        """
        # Get image full dimensions
        real_width, real_height = self.image.size
        real_dimensions = itertools.cycle((real_width, real_height))

        dimensions = []
        percentage = False
        if coordinates.startswith('pct:'):
            for coordinate in coordinates.split(':')[1].split(','):
                dimensions.append(float(coordinate))
            percentage = True
        else:
            for coordinate in coordinates.split(','):
                dimensions.append(int(coordinate))

        # First check if it has 4 coordinates x,y,w,h
        dimensions_length = len(dimensions)
        if dimensions_length != 4:
            raise MultimediaImageCropError(
                "Must have 4 dimensions {0} has been given".
                format(dimensions_length))

        # Make sure that there is not any negative dimension
        if any(coordinate < 0 for coordinate in dimensions):
            raise MultimediaImageCropError(
                "Dimensions cannot be negative {0} has been given".
                format(dimensions)
            )

        if percentage:
            if any(coordinate > 100.0 for coordinate in dimensions):
                raise MultimediaImageCropError(
                    "Dimensions could not be grater than 100%")

            # Calculate the dimensions
            start_x, start_y, width, height = [
                int(
                    math.floor(
                        self.percent_to_number(dimension) *
                        next(real_dimensions)
                    )
                ) for dimension in dimensions
            ]
        else:
            start_x, start_y, width, height = dimensions

        # Check if any of the requested axis is outside of image borders
        if any(axis > next(real_dimensions) for axis in (start_x, start_y)):
            raise MultimediaImageCropError(
                "Outside of image borders {0},{1}".
                format(real_width, real_height)
            )

        # Calculate the final dimensions
        max_x = start_x + width
        max_y = start_y + height
        # Check if the final width is bigger than the the real image width
        if max_x > real_width:
            max_x = real_width

        # Check if the final height is bigger than the the real image height
        if max_y > real_height:
            max_y = real_height

        self.image = self.image.crop((start_x, start_y, max_x, max_y))

    def rotate(self, degrees, mirror=False):
        """Rotate the image by given degrees.

        :param float degrees: The degrees, should be in range of [0, 360]
        :param bool mirror: Flip image from left to right
        """
        transforms = {
            '90': Image.ROTATE_90,
            '180': Image.ROTATE_180,
            '270': Image.ROTATE_270,
            'mirror': Image.FLIP_LEFT_RIGHT,
        }

        # Check if we have the right degrees
        if not 0.0 <= float(degrees) <= 360.0:
            raise MultimediaImageRotateError(
                "Degrees must be between 0 and 360, {0} has been given".
                format(degrees)
            )

        if str(degrees) in transforms.keys():
            self.image = self.image.transpose(transforms.get(str(degrees)))
        else:
            # transparent background if degrees not multiple of 90
            self.image = self.image.convert('RGBA')
            self.image = self.image.rotate(float(degrees), expand=0)

        if mirror:
            self.image = self.image.transpose(transforms.get('mirror'))

    def quality(self, quality):
        """Change the image format.

        :param str quality: The image quality should be in (default, grey,
                        bitonal, color)

        .. note::

            The library supports transformations between each supported
            mode and the "L" and "RGB" modes. To convert between other
            modes, you may have to use an intermediate image (typically
            an "RGB" image).

        """
        qualities = current_app.config['IIIF_QUALITIES']
        if quality not in qualities:
            raise MultimediaImageQualityError(
                ("{0} is not supported, please select one of the"
                 " valid qualities: {1}").format(quality, qualities)
            )

        qualities_by_code = zip(qualities,
                                current_app.config['IIIF_CONVERTERS'])

        if quality not in ('default', 'color'):
            # Convert image to RGB read the note
            if self.image.mode != "RGBA":
                self.image = self.image.convert('RGBA')

            code = [quality_code[1] for quality_code in qualities_by_code
                    if quality_code[0] == quality][0]

            self.image = self.image.convert(code)

    def size(self):
        """Return the current image size.

        :return: the image size
        """
        return self.image.size

    def save(self, path, image_format="jpeg", quality=90):
        """Store the image to the specific path.

        :param str path: absolute path
        :param str image_format: (gif, jpeg, pdf, png, tif)
        :param int quality: The image quality; [1, 100]

        .. note::

            `image_format` = jpg will not be recognized by :py:mod:`PIL.Image`
            and it will be changed to jpeg.

        """
        # transform `image_format` is lower case and not equals to jpg
        cleaned_image_format = self._prepare_for_output(image_format)
        self.image.save(path, cleaned_image_format, quality=quality)

    def serve(self, image_format="png", quality=90):
        """Return a BytesIO object to easily serve it thought HTTTP.

        :param str image_format: (gif, jpeg, pdf, png, tif)
        :param int quality: The image quality; [1, 100]

        .. note::

            `image_format` = jpg will not be recognized by
            :py:mod:`PIL.Image` and it will be changed to jpeg.

        """
        image_buffer = BytesIO()
        # transform `image_format` is lower case and not equals to jpg
        cleaned_image_format = self._prepare_for_output(image_format)
        save_kwargs = dict(quality=quality)

        if self.image.format == 'GIF':
            save_kwargs.update(save_all=True)

        self.image.save(image_buffer, cleaned_image_format, **save_kwargs)
        image_buffer.seek(0)

        return image_buffer

    def _prepare_for_output(self, requested_format):
        """Help validate output format.

        :param str requested_format: The image output format

        .. note::

            pdf and jpeg format can't be saved as `RBGA` so image needs to be
            converted to `RGB` mode.

        """
        image_format = self.sanitize_format_name(requested_format)
        format_keys = current_app.config['IIIF_FORMATS'].keys()

        if image_format not in format_keys:
            raise MultimediaImageFormatError(
                ("{0} does not supported, please select on of the valid"
                 " formats: {1}").format(requested_format, format_keys)
            )

        # If the the `requested_format` is pdf or jpeg force mode to RGB
        if image_format in ("pdf", "jpeg"):
            self.image = self.image.convert('RGB')

        return image_format

    @staticmethod
    def reduce_by(nominally, dominator):
        """Calculate the ratio."""
        return float(nominally) / float(dominator)

    @staticmethod
    def percent_to_number(number):
        """Calculate the percentage."""
        return float(number) / 100.0

    @staticmethod
    def sanitize_format_name(value):
        """Lowercase formats and make sure that jpg is written as jpeg."""
        return value.lower().replace("jpg", "jpeg")


class IIIFImageAPIWrapper(MultimediaImage):
    """IIIF Image API Wrapper."""

    @staticmethod
    def validate_api(**kwargs):
        """Validate IIIF Image API.

        Example to validate the IIIF API:

        .. code:: python

            from flask_iiif.api import IIIFImageAPIWrapper

            IIIFImageAPIWrapper.validate_api(
                version=version,
                region=region,
                size=size,
                rotation=rotation,
                quality=quality,
                image_format=image_format
            )

        .. note::

            If the version is not specified it will fallback to version 2.0.

        """
        # Get the api version
        version = kwargs.get('version', 'v2')
        # Get the validations and ignore cases
        cases = current_app.config['IIIF_VALIDATIONS'].get(version)
        for key in cases.keys():
            # If the parameter don't match with iiif casess
            if not re.search(
                    cases.get(key, {}).get('validate', ''), kwargs.get(key)
            ):
                raise IIIFValidatorError(
                    ("value: `{0}` for parameter: `{1}` is not supported").
                    format(kwargs.get(key), key)
                )

    def apply_api(self, **kwargs):
        """Apply the IIIF API to the image.

        Example to apply the IIIF API:

        .. code:: python

            from flask_iiif.api import IIIFImageAPIWrapper

            image = IIIFImageAPIWrapper.from_file(path)

            image.apply_api(
                version=version,
                region=region,
                size=size,
                rotation=rotation,
                quality=quality
            )

        .. note::

            * If the version is not specified it will fallback to version 2.0.
            * Please note the :func:`validate_api` should be run before
              :func:`apply_api`.

        """
        # Get the api version
        version = kwargs.get('version', 'v2')
        # Get the validations and ignore cases
        cases = current_app.config['IIIF_VALIDATIONS'].get(version)
        # Set the apply order
        order = 'region', 'size', 'rotation', 'quality'
        # Set the functions to be applied
        tools = {
            "region": self.apply_region,
            "size": self.apply_size,
            "rotation": self.apply_rotate,
            "quality": self.apply_quality
        }

        for key in order:
            # Ignore if has the ignore value for the specific key
            if kwargs.get(key) != cases.get(key, {}).get('ignore'):
                tools.get(key)(kwargs.get(key))

    def apply_region(self, value):
        """IIIF apply crop.

        Apply :func:`~flask_iiif.api.MultimediaImage.crop`.
        """
        self.crop(value)

    def apply_size(self, value):
        """IIIF apply resize.

        Apply :func:`~flask_iiif.api.MultimediaImage.resize`.
        """
        self.resize(value)

    def apply_rotate(self, value):
        """IIIF apply rotate.

        Apply :func:`~flask_iiif.api.MultimediaImage.rotate`.
        """
        mirror = False
        degrees = value
        if value.startswith('!'):
            mirror = True
            degrees = value[1:]
        self.rotate(degrees, mirror=mirror)

    def apply_quality(self, value):
        """IIIF apply quality.

        Apply :func:`~flask_iiif.api.MultimediaImage.quality`.
        """
        self.quality(value)

    @classmethod
    def open_image(cls, source):
        """Create an :class:`~flask_iiif.api.MultimediaImage` instance.

        :param str source: The image image string
        :type source: `BytesIO` object
        :param str source_type: the type of ``data``
        :returns: a :class:`~flask_iiif.api.MultimediaImage`
                  instance
        """
        try:
            image = Image.open(source)
        except (AttributeError, IOError):
            raise MultimediaImageNotFound(
                "The requested image cannot be opened"
            )
        return cls(image)



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015, 2016 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Flask-IIIF utilities."""
import shutil
import tempfile
from os.path import dirname, join

from flask import abort, url_for
from PIL import Image, ImageSequence

__all__ = ('iiif_image_url', )


def iiif_image_url(**kwargs):
    """Generate a `IIIF API` image url.

    :returns: `IIIF API` image url
    :rtype: str

    .. code:: html

        <img
            src="{{ iiif_image_url(uuid="id", size="200,200") }}"
            alt="Title"
        />

    .. note::

        If any IIIF parameter missing it will fall back to default:
        * `image_format = png`
        * `quality = default`
        * `region = full`
        * `rotation = 0`
        * `size = full`
        * `version = v2`
    """
    try:
        assert kwargs.get('uuid') is not None
    except AssertionError:
        abort(404)
    else:
        url_for_args = {k: v for k, v in kwargs.items() if k.startswith('_')}

        return url_for(
            'iiifimageapi',
            image_format=kwargs.get('image_format', 'png'),
            quality=kwargs.get('quality', 'default'),
            region=kwargs.get('region', 'full'),
            rotation=kwargs.get('rotation', 0),
            size=kwargs.get('size', 'full'),
            uuid=kwargs.get('uuid'),
            version=kwargs.get('version', 'v2'),
            **url_for_args
        )


def create_gif_from_frames(frames, duration=500, loop=0):
    """Create a GIF image.

    :param frames: the sequence of frames that resulting GIF should contain
    :param duration: the duration of each frame (in milliseconds)
    :param loop: the number of iterations of the frames (0 for infinity)
    :returns: GIF image
    :rtype: PIL.Image

    .. note:: Uses ``tempfile``, as PIL allows GIF creation only on ``save``.
    """
    # Save GIF to temporary file
    tmp = tempfile.mkdtemp(dir=dirname(__file__))
    tmp_file = join(tmp, 'temp.gif')

    head, tail = frames[0], frames[1:]
    head.save(tmp_file, 'GIF',
              save_all=True,
              append_images=tail,
              duration=duration,
              loop=loop)

    gif_image = Image.open(tmp_file)
    assert gif_image.is_animated

    # Cleanup temporary file
    shutil.rmtree(tmp)

    return gif_image


def resize_gif(image, size, resample):
    """Resize a GIF image.

    :param image: the original GIF image
    :param size: the dimensions to resize to
    :param resample: the method of resampling
    :returns: resized GIF image
    :rtype: PIL.Image
    """
    return create_gif_from_frames([frame.resize(size, resample=resample)
                                  for frame in ImageSequence.Iterator(image)])


def resize_with_background_gif(image, size, resample, demand_w=0, demand_h=0):
    """Resize a GIF image and fills the missing pixels with black background.

    :param image: the original GIF image
    :param size: the dimensions to resize to
    :param resample: the method of resampling
    :param demand_w: demanded height for thumbnail
    :param demand_h: demanded width for thumbnail
    :returns: resized GIF image
    :rtype: PIL.Image
    """
    return create_gif_from_frames([fill_background(
        frame.resize(size, resample=resample), demand_w, demand_h)
        for frame in ImageSequence.Iterator(image)])


def fill_background(image, demand_width, demand_height):
    """Fill the background with black (in case of image too small).

    :param image: image which does not fit into the window
    :param demand_width: demanded thumbnail window width
    :param demand_height: demanded window height
    :returns: image of requested size, filled with black background
    :rtype: PIL.Image
    """
    background = Image.new('RGB', (demand_width, demand_height))
    offset_x, offset_y = 0, 0
    w, h = image.size
    if w < demand_width:
        # set the image in the middle of x axis
        offset_x = max(1, int((demand_width - w) / 2))
    if h < demand_height:
        # set the image in the middle of y axis
        offset_y = max(1, int((demand_height - h) / 2))
    background.paste(image, (offset_x, offset_y))
    return background



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2015 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Test REST API."""

from io import BytesIO

from flask import url_for
from PIL import Image
from werkzeug.utils import secure_filename

from .helpers import IIIFTestCase


class TestRestAPI(IIIFTestCase):

    """Test signals and decorators."""

    def test_api_base(self):
        """Test API Base."""
        data = dict(
            uuid="valid:id",
            version="v2"
        )
        get_the_response = self.get(
            'iiifimagebase',
            urlargs=data
        )
        self.assertEqual(get_the_response.status_code, 303)

    def test_api_info(self):
        """Test API Info not found case."""
        from flask import jsonify
        id_v1 = url_for(
            'iiifimagebase', uuid=u"valid:id-üni", version="v1", _external=True
        )
        id_v2 = url_for(
            'iiifimagebase', uuid=u"valid:id-üni", version="v2", _external=True
        )

        expected = {
            "v1": {
                "@context": (
                    "http://library.stanford.edu/iiif/"
                    "image-api/1.1/context.json"
                ),
                "@id": id_v1,
                "width": 1280,
                "height": 1024,
                "profile": (
                    "http://library.stanford.edu/iiif/image-api/compliance"
                    ".html#level1"
                ),
                "tile_width": 256,
                "tile_height": 256,
                "scale_factors": [
                    1, 2, 4, 8, 16, 32, 64
                ]
            },
            "v2": {
                "@context": "http://iiif.io/api/image/2/context.json",
                "@id": id_v2,
                "protocol": "http://iiif.io/api/image",
                "width": 1280,
                "height": 1024,
                "tiles": [
                    {
                        "width": 256,
                        "scaleFactors": [
                            1, 2, 4, 8, 16, 32, 64
                        ]
                    }
                ],
                "profile": [
                    "http://iiif.io/api/image/2/level2.json",
                ]
            }
        }
        get_the_response = self.get(
            'iiifimageinfo',
            urlargs=dict(
                uuid=u'valid:id-üni',
                version='v2',
            )
        )
        self.assert200(get_the_response)
        get_the_response = self.get(
            'iiifimageinfo',
            urlargs=dict(
                uuid=u'valid:id-üni',
                version='v2',
            )
        )
        self.assertEqual(
            jsonify(expected.get('v2')).data, get_the_response.data
        )

        get_the_response = self.get(
            'iiifimageinfo',
            urlargs=dict(
                uuid=u'valid:id-üni',
                version='v1',
            )
        )
        self.assert200(get_the_response)
        self.assertEqual(
            jsonify(expected.get('v1')).data, get_the_response.data
        )

    def test_api_info_not_found(self):
        """Test API Info."""
        get_the_response = self.get(
            'iiifimageinfo',
            urlargs=dict(
                uuid='notfound',
                version='v2',
            )
        )
        self.assert404(get_the_response)

    def test_api_not_found(self):
        """Test API not found case."""
        get_the_response = self.get(
            'iiifimageapi',
            urlargs=dict(
                uuid='notfound',
                version='v2',
                region='full',
                size='full',
                rotation='0',
                quality='default',
                image_format='png'
            )
        )
        self.assert404(get_the_response)

    def test_api_internal_server_error(self):
        """Test API internal server error case."""
        get_the_response = self.get(
            'iiifimageapi',
            urlargs=dict(
                uuid=u'valid:id-üni',
                version='v2',
                region='full',
                size='full',
                rotation='2220',
                quality='default',
                image_format='png'
            )
        )
        self.assert500(get_the_response)

    def test_api_iiif_validation_error(self):
        """Test API iiif validation case."""
        get_the_response = self.get(
            'iiifimageapi',
            urlargs=dict(
                uuid='valid:id-üni',
                version='v1',
                region='200',
                size='full',
                rotation='2220',
                quality='default',
                image_format='png'
            )
        )
        self.assert400(get_the_response)

    def test_api_stream_image(self):
        """Test API stream image."""
        tmp_file = BytesIO()
        # create a new image
        image = Image.new("RGBA", (1280, 1024), (255, 0, 0, 0))
        image.save(tmp_file, 'png')
        tmp_file.seek(0)

        get_the_response = self.get(
            'iiifimageapi',
            urlargs=dict(
                uuid=u'valid:id-üni',
                version='v2',
                region='full',
                size='full',
                rotation='0',
                quality='default',
                image_format='png'
            )
        )

        self.assertEqual(
            get_the_response.data,
            tmp_file.getvalue()
        )

        urlargs = dict(
            uuid=u'valid:id-üni',
            version='v2',
            region='200,200,200,200',
            size='300,300',
            rotation='!50',
            quality='color',
            image_format='pdf',
        )

        get_the_response = self.get(
            'iiifimageapi',
            urlargs=urlargs,
        )
        self.assert200(get_the_response)

        default_name = u'{name}-200200200200-300300-color-50.pdf'.format(
            name=secure_filename(urlargs['uuid'])
        )
        for dl, name in (('', default_name), ('1', default_name),
                         ('foo.pdf', 'foo.pdf')):
            urlargs['dl'] = dl
            get_the_response = self.get(
                'iiifimageapi',
                urlargs=urlargs,
            )
            self.assert200(get_the_response)
            self.assertEqual(
                get_the_response.headers['Content-Disposition'],
                'attachment; filename={name}'.format(name=name)
            )

    def test_api_decorator(self):
        """Test API decorator."""
        get_the_response = self.get(
            'iiifimageapi',
            urlargs=dict(
                uuid='valid:decorator:id',
                version='v2',
                region='full',
                size='full',
                rotation='0',
                quality='default',
                image_format='png'
            )
        )
        self.assert403(get_the_response)

    def test_api_abort_all_methods_except_get(self):
        """Abort all methods but GET."""
        data = dict(
            uuid='valid:id-üni',
            version='v2',
            region='full',
            size='full',
            rotation='0',
            quality='default',
            image_format='png'
        )
        get_the_response = self.post(
            'iiifimageapi',
            urlargs=data
        )
        self.assert405(get_the_response)

        get_the_response = self.put(
            'iiifimageapi',
            urlargs=data
        )
        self.assert405(get_the_response)

        get_the_response = self.delete(
            'iiifimageapi',
            urlargs=data
        )
        self.assert405(get_the_response)

        get_the_response = self.patch(
            'iiifimageapi',
            urlargs=data
        )
        self.assert405(get_the_response)



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015, 2016, 2017 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""IIIF configuration.

.. py:data:: IIIF_CACHE_HANDLER

    Add the preferred cache adaptor.

    .. seealso:: :py:class:`~flask_iiif.cache.cache.ImageCache`

.. py:data:: IIIF_CACHE_TIME

    How much time the image would be cached.

.. py:data:: IIIF_QUALITIES

    The supported image qualities.

    .. seealso::

            `IIIF Image API
            <http://iiif.io/api/image/2.0/#quality>`_

.. py:data:: IIIF_CONVERTERS

    The supported image converters.

.. py:data: IIIF_FORMATS

    The supported image formats with their MIME type.

.. py:data:: IIIF_VALIDATIONS

    The IIIF Image API validation.

    .. seealso::

        `IIIF Image API v1
        <http://iiif.io/api/image/1.1/>`_ and
        `IIIF Image API v2
        <http://iiif.io/api/image/2.0/>`_

.. py:data:: IIIF_API_INFO_RESPONSE_SKELETON

    Information request document for the image.

    .. seealso::
        `IIIF Image API v1 Information request
        <http://iiif.io/api/image/1.1/#information-request>`_ and
        `IIIF Image API v2 Information request
        <http://iiif.io/api/image/2.0/#information-request>`_

"""
# Resampling method
IIIF_RESIZE_RESAMPLE = 'PIL.Image:BICUBIC'

# Cache handler
IIIF_CACHE_HANDLER = 'flask_iiif.cache.simple:ImageSimpleCache'

# Cache duration
# 60 seconds * 60 minutes (1 hour) * 24 (24 hours) * 2 (2 days) = 172800 secs
# 60 seconds * 60 (1 hour) * 24 (1 day) * 2 (2 days)
IIIF_CACHE_TIME = 60 * 60 * 24 * 2

# Redis URL Cache
IIIF_CACHE_REDIS_URL = 'redis://localhost:6379/0'

# Supported qualities
IIIF_QUALITIES = (
    'default', 'grey', 'bitonal', 'color', 'native'
)
# Suported coverters
IIIF_CONVERTERS = '', 'L', '1', '', ''

# Supported multimedia image API formats
IIIF_FORMATS = {
    'gif': 'image/gif',
    'jpeg': 'image/jpeg',
    'jpg': 'image/jpeg',
    'pdf': 'application/pdf',
    'png': 'image/png',
    'tif': 'image/tiff',
}

# Regular expressions to validate each parameter
IIIF_VALIDATIONS = {
    "v1": {
        "region": {
            "ignore": "full",
            "validate": "(^full|(pct:)?([\d.]+,){3}([\d.]+))",
        },
        "size": {
            "ignore": "full",
            "validate": ("(^full|[\d.]+,|,[\d.]+|pct:[\d.]+|[\d.]+,"
                         "[\d.]+|![\d.]+,[\d.]+)")
        },
        "rotation": {
            "ignore": "0",
            "validate": "^[\d.]+$"
        },
        "quality": {
            "ignore": "default",
            "validate": "(native|color|grey|bitonal)"
        },
        "image_format": {
            "ignore": "",
            "validate": "(jpg|png|gif|jp2|jpeg|pdf|tif)"
        }
    },
    "v2": {
        "region": {
            "ignore": "full",
            "validate": "(^full|(pct:)?([\d.]+,){3}([\d.]+))",
        },
        "size": {
            "ignore": "full",
            "validate": ("(^full|[\d.]+,|,[\d.]+|pct:[\d.]+|[\d.]+,"
                         "[\d.]+|![\d.]+,[\d.]+)")
        },
        "rotation": {
            "ignore": "0",
            "validate": "^!?[\d.]+$"
        },
        "quality": {
            "ignore": "default",
            "validate": "(default|color|grey|bitonal)"
        },
        "image_format": {
            "ignore": "",
            "validate": "(jpg|png|gif|jp2|jpeg|pdf|tif)"
        }
    }
}

# Qualities per image mode
IIIF_MODE = {
    '1': ['default', 'bitonal'],
    'L': ['default', 'gray', 'bitonal'],
    'P': ['default', 'gray', 'bitonal'],
    'RGB': ['default', 'color', 'gray', 'bitonal'],
    'RGBA': ['default', 'color', 'gray', 'bitonal'],
    'CMYK': ['default', 'color', 'gray', 'bitonal'],
    'YCbCr': ['default', 'color', 'gray', 'bitonal'],
    'I': ['default', 'color', 'gray', 'bitonal'],
    'F': ['default', 'color', 'gray', 'bitonal']
}

# API Info
IIIF_API_INFO_RESPONSE_SKELETON = {
    "v1": {
        "@context": (
            "http://library.stanford.edu/iiif/image-api/1.1/context.json"
        ),
        "@id": "",
        "width": "",
        "height": "",
        "profile": (
            "http://library.stanford.edu/iiif/image-api/compliance.html#level1"
        ),
        "tile_width": 256,
        "tile_height": 256,
        "scale_factors": [
            1, 2, 4, 8, 16, 32, 64
        ]
    },
    "v2": {
        "@context": "http://iiif.io/api/image/2/context.json",
        "@id": "",
        "protocol": "http://iiif.io/api/image",
        "width": "",
        "height": "",
        "tiles": [
            {
                "width": 256,
                "scaleFactors": [
                    1, 2, 4, 8, 16, 32, 64
                ]
            }
        ],
        "profile": [
            "http://iiif.io/api/image/2/level2.json",
        ]
    }
}



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015, 2016, 2017 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Multimedia IIIF Image API."""

from io import BytesIO

from flask import current_app, jsonify, redirect, request, send_file, url_for
from flask_restful import Resource
from flask_restful.utils import cors
from werkzeug import LocalProxy
from werkzeug.utils import secure_filename

from .api import IIIFImageAPIWrapper
from .decorators import api_decorator, error_handler
from .signals import iiif_after_info_request, iiif_after_process_request, \
    iiif_before_info_request, iiif_before_process_request

current_iiif = LocalProxy(lambda: current_app.extensions['iiif'])


class IIIFImageBase(Resource):
    """IIIF Image Base."""

    def get(self, version, uuid):
        """Get IIIF Image Base.

        .. note::

            It will redirect to ``iiifimageinfo`` endpoint with status code
            303.
        """
        return redirect(
            url_for('iiifimageinfo', version=version, uuid=uuid), code=303
        )


class IIIFImageInfo(Resource):
    """IIIF Image Info."""

    method_decorators = [
        error_handler,
        api_decorator,
    ]

    @cors.crossdomain(origin='*', methods='GET')
    def get(self, version, uuid):
        """Get IIIF Image Info."""
        # Trigger event before proccess the api request
        iiif_before_info_request.send(self, version=version, uuid=uuid)

        # Check if its cached
        cache_handler = current_iiif.cache()

        # build the image key
        key = u"iiif:info:{0}/{1}".format(
            version, uuid
        ).encode('utf8')

        # Check if its cached
        cached = cache_handler.get(key)

        # If the image size is cached loaded from cache
        if cached:
            width, height = map(int, cached.split(','))
        else:
            data = current_iiif.uuid_to_image_opener(uuid)
            image = IIIFImageAPIWrapper.open_image(data)
            width, height = image.size()
            cache_handler.set(key, "{0},{1}".format(width, height))

        data = current_app.config['IIIF_API_INFO_RESPONSE_SKELETON'][version]

        base_uri = url_for(
            'iiifimagebase',
            uuid=uuid,
            version=version,
            _external=True
        )
        data["@id"] = base_uri
        data["width"] = width
        data["height"] = height

        # Trigger event after proccess the api request
        iiif_after_info_request.send(self, **data)

        return jsonify(data)


class IIIFImageAPI(Resource):
    """IIIF API Implementation.

    .. note::

        * IIF IMAGE API v1.0
            * For more infos please visit <http://iiif.io/api/image/>.
        * IIIF Image API v2.0
            * For more infos please visit <http://iiif.io/api/image/2.0/>.
        * The API works only for GET requests
        * The image process must follow strictly the following workflow:

           * Region
           * Size
           * Rotation
           * Quality
           * Format

    """

    method_decorators = [
        error_handler,
        api_decorator,
    ]

    def get(self, version, uuid, region, size, rotation, quality,
            image_format):
        """Run IIIF Image API workflow."""
        api_parameters = dict(
            version=version,
            uuid=uuid,
            region=region,
            size=size,
            rotation=rotation,
            quality=quality,
            image_format=image_format
        )
        # Trigger event before proccess the api request
        iiif_before_process_request.send(self, **api_parameters)

        # Validate IIIF parameters
        IIIFImageAPIWrapper.validate_api(**api_parameters)

        cache_handler = current_iiif.cache()

        # build the image key
        key = u'iiif:{0}/{1}/{2}/{3}/{4}.{5}'.format(
            uuid, region, size, quality, rotation, image_format
        ).encode('utf8')

        # Check if its cached
        cached = cache_handler.get(key)

        # If the image is cached loaded from cache
        if cached:
            to_serve = BytesIO(cached)
            to_serve.seek(0)
        # Otherwise create the image
        else:
            data = current_iiif.uuid_to_image_opener(uuid)
            image = IIIFImageAPIWrapper.open_image(data)

            image.apply_api(
                version=version,
                region=region,
                size=size,
                rotation=rotation,
                quality=quality
            )

            # prepare image to be serve
            to_serve = image.serve(image_format=image_format)
            # to_serve = image.serve(image_format=image_format)
            cache_handler.set(key, to_serve.getvalue())

        # decide the mime_type from the requested image_format
        mimetype = current_app.config['IIIF_FORMATS'].get(
            image_format, 'image/jpeg'
        )
        # Built the after request parameters
        api_after_request_parameters = dict(
            mimetype=mimetype,
            image=to_serve
        )

        # Trigger event after proccess the api request
        iiif_after_process_request.send(self, **api_after_request_parameters)

        send_file_kwargs = {'mimetype': mimetype}
        if 'dl' in request.args:
            filename = secure_filename(request.args.get('dl', ''))
            if filename.lower() in {'', '1', 'true'}:
                filename = u'{0}-{1}-{2}-{3}-{4}.{5}'.format(
                    uuid, region, size, quality, rotation, image_format
                )
            send_file_kwargs.update(
                as_attachment=True,
                attachment_filename=secure_filename(filename),
            )

        return send_file(to_serve, **send_file_kwargs)



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2015, 2016, 2017 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Signals triggered on Flask-IIIF API request.

.. note::

    The signals are triggered before process, after the validation and after
    the process of the request.

.. code-block:: python

    from flask import current_app

    from flask_iiif.signals import (
        iiif_after_process_request, iiif_before_process_request
    )

    def on_before_process_request(
            current_object_state, version='', uuid='', region='', size='',
            rotate='', quality='', image_format=''):
        # Do something before process the request

    iiif_before_process_request.connect_via(on_before_process_request)

    def on_after_process_request(
            current_object_state, mimetype='', image=''):
        # Do something after proccess the request

    iiif_after_process_request.connect(on_after_process_request)

"""

from blinker import Namespace

_signals = Namespace()

# Before request
iiif_before_process_request = _signals.signal('iiif-before-process-request')
# After request
iiif_after_process_request = _signals.signal('iiif-after-process-request')
# Before info.json request
iiif_before_info_request = _signals.signal('iiif-before-info-request')
# After info.json request
iiif_after_info_request = _signals.signal('iiif-after-info-request')



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015, 2017 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Abstract simple cache definition.

All cache adaptors must at least implement
:func:`~flask_iiif.cache.cache.ImageCache.get` and
:func:`~flask_iiif.cache.cache.ImageCache.set` methods.
"""

from flask import current_app
from werkzeug import cached_property


class ImageCache(object):
    """Abstract cache layer."""

    def __init__(self):
        """Initialize the cache."""

    @cached_property
    def timeout(self):
        """Return default timeout from config."""
        return current_app.config['IIIF_CACHE_TIME']

    def get(self, key):
        """Return the key value.

        :param key: the object's key
        """

    def set(self, key, value, timeout=None):
        """Cache the object.

        :param key: the object's key
        :param value: the stored object
        :type value: :class:`StringIO.StringIO` object
        :param timeout: the cache timeout in seconds
        """

    def delete(self, key):
        """Delete the specific key."""

    def flush(self):
        """Flush the cache."""



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015, 2017 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Version information for Flask-IIIF.

This file is imported by ``flask_iiif.__init__``, and parsed by
``setup.py`` as well as ``docs/conf.py``.
"""

# Do not change the format of this next line. Doing so risks breaking
# setup.py and docs/conf.py

__version__ = "0.5.0"



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015, 2016, 2017 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Multimedia and IIIF Image APIs.

Flask-IIIF is initialized like this:

Initialization of the extension:

>>> from flask import Flask
>>> from flask_iiif import IIIF
>>> app = Flask('myapp')
>>> ext = IIIF(app=app)

or alternatively using the factory pattern:

>>> app = Flask('myapp')
>>> ext = IIIF()
>>> ext.init_app(app)
"""

from __future__ import absolute_import

from flask import current_app

from six import string_types

from werkzeug.urls import url_join
from werkzeug.utils import import_string

from . import config
from .utils import iiif_image_url
from .version import __version__


class IIIF(object):
    """Flask extension implementation."""

    def __init__(self, app=None):
        """Initialize login callback."""
        self.uuid_to_image_opener = None
        self.api_decorator_callback = None
        if app is not None:
            self.init_app(app)

    @staticmethod
    def cache():
        """Return the cache handler.

        .. note::

            You can create your own cache handler by change the
            :py:attr:`~flask_iiif.config.IIIF_CACHE_HANDLER`. More infos
            could be found in :py:mod:`~flask_iiif.cache.cache`.
        """
        from .cache.cache import ImageCache
        handler = current_app.config['IIIF_CACHE_HANDLER']
        if isinstance(handler, string_types):
            handler = import_string(handler)
        if callable(handler):
            handler = handler()
        assert isinstance(handler, ImageCache)
        return handler

    def init_app(self, app):
        """Initialize a Flask application."""
        self.app = app
        # Follow the Flask guidelines on usage of app.extensions
        if not hasattr(app, 'extensions'):
            app.extensions = {}
        if 'iiif' in app.extensions:
            raise RuntimeError("Flask application already initialized")
        app.extensions['iiif'] = self

        # Set default configuration
        for k in dir(config):
            if k.startswith('IIIF_'):
                self.app.config.setdefault(k, getattr(config, k))

        # Register context processors
        if hasattr(app, 'add_template_global'):
            app.add_template_global(iiif_image_url)
        else:
            ctx = dict(
                iiif_image_url=iiif_image_url
            )
            app.context_processor(lambda: ctx)

    def init_restful(self, api, prefix='/api/multimedia/image/'):
        """Set up the urls.

        :param str prefix: the url perfix

        .. note::

            In IIIF Image API the Image Request URI Syntax must following

                ``{scheme}://{server}{/prefix}/{identifier}/
                    {region}/{size}/{rotation}/{quality}.{format}``

            pattern, the default prefix is ``/api/multimedia/image`` but
            this can be changes by changing the ``prefix`` paremeter. The
            ``prefix`` MUST always start and end with `/`

        .. seealso::
            `IIIF IMAGE API URI Syntax
            <http://iiif.io/api/image/2.0/#uri-syntax>`
        """
        from .restful import IIIFImageAPI, IIIFImageInfo, IIIFImageBase

        if not prefix.startswith('/') or not prefix.endswith('/'):
            raise RuntimeError(
                "The `prefix` must always start and end with `/`"
            )

        api.add_resource(
            IIIFImageAPI,
            url_join(
                prefix,
                (
                    "<string:version>/<string:uuid>/"
                    "<string:region>/<string:size>/<string:rotation>/"
                    "<string:quality>.<string:image_format>"
                )
            )
        )

        api.add_resource(
            IIIFImageInfo,
            url_join(
                prefix,
                "<string:version>/<string:uuid>/info.json"
            )
        )
        api.add_resource(
            IIIFImageBase,
            url_join(
                prefix,
                "<string:version>/<string:uuid>"
            )
        )

    def uuid_to_image_opener_handler(self, callback):
        """Set the callback for the ``uuid`` to ``image`` convertion.

        .. note:

            The supported file type is either ``fullpath`` or ``bytestream``
            object anything else will raise
            :class:`~flask_iiif.errors.MultimediaImageNotFound`

        .. code-block:: python

            def uuid_to_path(uuid):
                # do something magical

            iiif.uuid_to_image_opener_handler(uuid_to_path)
        """
        self.uuid_to_image_opener = callback

    def api_decorator_handler(self, callback):
        """Protect API handler.

        .. code-block:: python

            def protect_api():
                return
            iiif.api_decorator_handler(protect_api)

        .. note::

            The API would be always decorated with ``api_decorator_handler``.
            If is not defined, it would just pass.
        """
        self.api_decorator_callback = callback

__all__ = ('IIIF', '__version__')



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2016, 2017 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Implements a Redis cache."""

from __future__ import absolute_import

from flask import current_app
from redis import StrictRedis
from werkzeug.contrib.cache import RedisCache

from .cache import ImageCache


class ImageRedisCache(ImageCache):
    """Redis image cache."""

    def __init__(self):
        """Initialize the cache."""
        super(ImageRedisCache, self).__init__()
        redis_url = current_app.config['IIIF_CACHE_REDIS_URL']
        self.cache = RedisCache(host=StrictRedis.from_url(redis_url))

    def get(self, key):
        """Return the key value.

        :param key: the object's key
        :return: the stored object
        :rtype: `BytesIO` object
        """
        return self.cache.get(key)

    def set(self, key, value, timeout=None):
        """Cache the object.

        :param key: the object's key
        :param value: the stored object
        :type value: `BytesIO` object
        :param timeout: the cache timeout in seconds
        """
        timeout = timeout if timeout else self.timeout
        self.cache.set(key, value, timeout=timeout)

    def delete(self, key):
        """Delete the specific key."""
        self.cache.delete(key)

    def flush(self):
        """Flush the cache."""
        self.cache.clear()



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2015 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Context Template Tests."""

from .helpers import IIIFTestCase


class TestContextTemplates(IIIFTestCase):

    """Context templates test case."""

    def test_context_tempalte_iiif_image_url(self):
        """Test context template."""
        from werkzeug.exceptions import NotFound
        from flask_iiif.utils import iiif_image_url

        with self.app.app_context():
            image_url_default = iiif_image_url(
                uuid=u"test-ünicode"
            )
            image_url_default_answer = (
                "/api/multimedia/image/v2/"
                "test-%C3%BCnicode/full/full/0/default.png"
            )
            self.assertEqual(
                image_url_default_answer,
                image_url_default
            )
            image_url_custom_answer = (
                "/api/multimedia/image/v1/"
                "test-%C3%BCnicode/full/full/180/default.jpg"
            )
            image_url_custom = iiif_image_url(
                uuid=u"test-ünicode",
                image_format="jpg",
                rotation=180,
                version="v1"
            )
            self.assertEqual(
                image_url_custom_answer,
                image_url_custom
            )
            self.assertRaises(
                NotFound,
                iiif_image_url,
                size="200,"
            )



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2015, 2016 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Flask-IIIF decorators."""

from functools import wraps

from flask import current_app
from flask_restful import abort
from werkzeug import LocalProxy

from .errors import IIIFValidatorError, MultimediaError, \
    MultimediaImageCropError, MultimediaImageFormatError, \
    MultimediaImageNotFound, MultimediaImageQualityError, \
    MultimediaImageResizeError, MultimediaImageRotateError

__all__ = ('api_decorator', 'error_handler', )

current_iiif = LocalProxy(lambda: current_app.extensions['iiif'])


def error_handler(f):
    """Error handler."""
    @wraps(f)
    def inner(*args, **kwargs):
        """Wrap the errors."""
        try:
            return f(*args, **kwargs)
        except (MultimediaImageCropError, MultimediaImageResizeError,
                MultimediaImageFormatError, MultimediaImageRotateError,
                MultimediaImageQualityError) as error:
            abort(500, message=error.message, code=500)
        except IIIFValidatorError as error:
            abort(400, message=error.message, code=400)
        except (MultimediaError, MultimediaImageNotFound) as error:
            abort(error.code, message=error.message, code=error.code)
    return inner


def api_decorator(f):
    """Decorate API method."""
    @wraps(f)
    def inner(*args, **kwargs):
        if current_iiif.api_decorator_callback:
            current_iiif.api_decorator_callback(*args, **kwargs)
        return f(*args, **kwargs)
    return inner



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2015, 2017 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Test REST API signals."""
import json

from .helpers import IIIFTestCase, signal_listener


class TestRestAPISignals(IIIFTestCase):

    """Test REST API signals."""

    def test_api_signals(self):
        """Test API signals."""
        from flask_iiif.signals import (
            iiif_after_info_request, iiif_after_process_request,
            iiif_before_info_request, iiif_before_process_request
        )

        data = dict(
            uuid='valid:id',
            version='v2',
            region='full',
            size='full',
            rotation='0',
            quality='default',
            image_format='png'
        )

        infodata = dict(
            version='v2',
            uuid='valid:id'
        )

        with signal_listener(iiif_before_process_request) as listener:
            self.get(
                'iiifimageapi',
                urlargs=data
            )
            results = listener.assert_signal()
            self.assertEqual(
                results[1],
                data
            )

        with signal_listener(iiif_after_process_request) as listener:
            response = self.get(
                'iiifimageapi',
                urlargs=data
            )
            results = listener.assert_signal()
            self.assertEqual(
                (results[1].get('mimetype'),
                 results[1].get('image').getvalue()),
                ('image/png', response.data)
            )

        with signal_listener(iiif_before_info_request) as listener:
            response = self.get(
                'iiifimageinfo',
                urlargs=infodata
            )
            results = listener.assert_signal()
            self.assertEqual(
                results[1],
                infodata
            )

        with signal_listener(iiif_after_info_request) as listener:
            response = self.get(
                'iiifimageinfo',
                urlargs=infodata
            )
            results = listener.assert_signal()
            self.assertEqual(
                results[1],
                json.loads(response.get_data(as_text=True))
            )



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015, 2016, 2017 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Flask-IIIF extension provides easy IIIF API standard integration."""

import os
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    """PyTest Test."""

    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        """Init pytest."""
        TestCommand.initialize_options(self)
        self.pytest_args = []
        try:
            from ConfigParser import ConfigParser
        except ImportError:
            from configparser import ConfigParser
        config = ConfigParser()
        config.read('pytest.ini')
        self.pytest_args = config.get('pytest', 'addopts').split(' ')

    def finalize_options(self):
        """Finalize pytest."""
        TestCommand.finalize_options(self)
        if hasattr(self, '_test_args'):
            self.test_suite = ''
        else:
            self.test_args = []
            self.test_suite = True

    def run_tests(self):
        """Run tests."""
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)

# Get the version string. Cannot be done with import!
g = {}
with open(os.path.join('flask_iiif', 'version.py'), 'rt') as fp:
    exec(fp.read(), g)
    version = g['__version__']

tests_require = [
    'flask-testing>=0.4.1',
    'check-manifest>=0.25',
    'coverage>=3.7,<4.0',
    'isort>=4.2.2',
    'pydocstyle>=2.0.0',
    'pytest-cache>=1.0',
    'pytest-cov>=1.8.0',
    'pytest-pep8>=1.0.6',
    'pytest>=2.6.1',
]

install_requires = [
    'Flask>=0.10.1',
    'Flask-RESTful>=0.2.12',
    'blinker>=1.4',
    'six>=1.7.2',
    'pillow>=3.4',
]

extra_require = {
    'docs': [
        'Sphinx>=1.4.2',
    ],
    'redis': [
        'redis>=2.10.5',
    ],
    'tests': tests_require,
}

extra_require['all'] = []
for reqs in extra_require.values():
    extra_require['all'].extend(reqs)

setup(
    name='Flask-IIIF',
    version=version,
    url='http://github.com/inveniosoftware/flask-iiif/',
    license='BSD',
    author='Invenio collaboration',
    author_email='info@inveniosoftware.org',
    description=__doc__,
    long_description=open('README.rst').read(),
    packages=['flask_iiif'],
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=install_requires,
    extras_require=extra_require,
    tests_require=tests_require,
    cmdclass={'test': PyTest},
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Framework :: Flask',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Development Status :: 5 - Production/Stable',
    ],
)



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015, 2016, 2017 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Implement a simple cache."""

from __future__ import absolute_import

from werkzeug.contrib.cache import SimpleCache

from .cache import ImageCache


class ImageSimpleCache(ImageCache):
    """Simple image cache."""

    def __init__(self):
        """Initialize the cache."""
        super(ImageSimpleCache, self).__init__()
        self.cache = SimpleCache()

    def get(self, key):
        """Return the key value.

        :param key: the object's key
        :return: the stored object
        :rtype: `BytesIO` object
        """
        return self.cache.get(key)

    def set(self, key, value, timeout=None):
        """Cache the object.

        :param key: the object's key
        :param value: the stored object
        :type value: `BytesIO` object
        :param timeout: the cache timeout in seconds
        """
        timeout = timeout if timeout else self.timeout
        self.cache.set(key, value, timeout)

    def delete(self, key):
        """Delete the specific key."""
        self.cache.delete(key)

    def flush(self):
        """Flush the cache."""
        self.cache.clear()



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Flask-IIIF test helpers."""

from contextlib import contextmanager
from io import BytesIO

from flask import Flask, abort, url_for
from flask_testing import TestCase


class IIIFTestCase(TestCase):

    """IIIF REST test case."""

    def create_app(self):
        """Create the app."""
        from flask_iiif import IIIF
        from flask_restful import Api
        from flask_iiif.cache.simple import ImageSimpleCache

        app = Flask(__name__)
        app.config['DEBUG'] = True
        app.config['TESTING'] = True
        app.config['SERVER_NAME'] = "shield.worker.node.1"
        app.config['SITE_URL'] = "http://shield.worker.node.1"
        app.config['IIIF_CACHE_HANDLER'] = ImageSimpleCache()
        app.logger.disabled = True

        api = Api(app=app)

        iiif = IIIF(app=app)

        iiif.uuid_to_image_opener_handler(self.create_image)

        def api_decorator_test(*args, **kwargs):
            if 'decorator' in kwargs.get('uuid'):
                abort(403)

        iiif.api_decorator_handler(api_decorator_test)

        iiif.init_restful(api)
        return app

    def get(self, *args, **kwargs):
        """Simulate a GET request."""
        return self.make_request(self.client.get, *args, **kwargs)

    def post(self, *args, **kwargs):
        """Simulate a POST request."""
        return self.make_request(self.client.post, *args, **kwargs)

    def put(self, *args, **kwargs):
        """Simulate a PUT request."""
        return self.make_request(self.client.put, *args, **kwargs)

    def delete(self, *args, **kwargs):
        """Simulate a DELETE request."""
        return self.make_request(self.client.delete, *args, **kwargs)

    def patch(self, *args, **kwargs):
        """Simulate a PATCH request."""
        return self.make_request(self.client.patch, *args, **kwargs)

    def make_request(self, client_func, endpoint, urlargs=None):
        """Simulate a request."""
        url = url_for(endpoint, **(urlargs or {}))
        response = client_func(
            url,
            base_url=self.app.config['SITE_URL']
        )
        return response

    def create_image(self, uuid):
        """Create a test image."""
        if uuid.startswith('valid'):
            from PIL import Image
            tmp_file = BytesIO()
            # create a new image
            image = Image.new("RGBA", (1280, 1024), (255, 0, 0, 0))
            image.save(tmp_file, 'png')
            tmp_file.seek(0)
            return tmp_file
        return ''


@contextmanager
def signal_listener(signal):
    """Context Manager that listen to signals.

    .. note::

        Checks if any signal were fired and returns the passed arguments.

    .. code-block:: python

        from somesignals import signal

        with signal_listener(signal) as listener:
            fire_the_signal()
            results = listener.assert_signal()
            # Assert the results here

    """
    class _Signal(object):
        def __init__(self):
            self.heard = []

        def add(self, *args, **kwargs):
            """Keep the signals."""
            self.heard.append((args, kwargs))

        def assert_signal(self):
            """Check signal."""
            if len(self.heard) == 0:
                raise AssertionError("No signals.")
            return self.heard[0][0], self.heard[0][1]

    signals = _Signal()
    signal.connect(signals.add)

    try:
        yield signals
    finally:
        signal.disconnect(signals.add)
